# -*- coding: utf-8 -*-
"""
Created on Wed Nov 18 22:04:35 2020

@author: weihao
"""

from prepextend.io.read import (
        decode_flow
        )


def flow_read(flow_file):

    flow = decode_flow(flow_file)
    inputs_outputs = flow.list_sources()

    response = {
        'raw_structure': flow.flow,
        'inputs_outputs': inputs_outputs
        }

    return response



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 18 22:04:35 2020

@author: weihao
"""

import json
import os
from os.path import join
import datetime
from pandas import read_csv
from math import ceil

from prepextend.management._config.config import config_set
from prepextend.management._notify.api import send_msg
from prepextend.common.api import list_files
from prepextend.io.api import flow_read
from prepextend.run.api import flow_run
from prepextend.map.api import flows_roadmap
from prepextend.map.api import construct_sourcesDF as construct_sources
from prepextend.common.api import path_split

# In[]:


class flow_manage:

    def __init__(self,
                 general_config,
                 version_assigned=None
                 ):

        self._config = config_set(general_config,
                                  version_assigned
                                  )
        # grt required configs
        self._versions_and_prep_scripts = self._config.prep_scripts_4_version_assigned
        self._retry_times = self._config.retry_times
        self._default_prep_script = self._config.default_prep_script
        self._credentials_file_suffix = self._config.credentials_file_suffix
        self._check_file_tags = self._config.check_file_tags
        self._log_save_dir = self._config.log_save_dir
        self._flow_pool = self._config.flow_pool
        self._version_assigned = self._config.version_assigned
        self._notify_channels = self._config.notify_channels

    def _assign_prep_script_byconfig(self, flow_file: str) -> str:

        flow_dir, flow_name = path_split(flow_file)

        flows_defined_version = self._version_assigned
        default_prep_script = self._default_prep_script
        versions_and_prep_scripts = self._versions_and_prep_scripts

        # check version
        if flow_file in flows_defined_version.keys():
            # use custom setting if exist
            flow_version = flows_defined_version[flow_file]
            print(f"{flow_name} has defined version: {flow_version}")
            if flow_version in versions_and_prep_scripts.keys():
                prep_script = versions_and_prep_scripts[flow_version]
            else:
                prep_script = default_prep_script
                print("but can't find the the prep_script in the config,"
                      "so use default_prep_script")
        else:
            prep_script = default_prep_script
            print(f"{flow_name} has no defined version,"
                  " use default_prep_script")

        return prep_script

    def _credential_file_byconfig(self, flow_file: str) -> str:

        flow_file_dir, flow_file_name = path_split(flow_file)

        credential_name = (os.path.splitext(flow_file_name)[0]
                           + self._credentials_file_suffix
                           )

        credential_file = join(flow_file_dir, credential_name)

        if os.path.isfile(credential_file) is False:
            return None
        else:
            return credential_file

    def _save_log(self, run_result: dict, log_save_dir: str):

        class MyEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, bytes):
                    return str(obj, encoding='utf-8', errors='ignore')
                return json.JSONEncoder.default(self, obj)

        start_time = run_result['start_time']
        start_f = datetime.datetime.fromtimestamp(
                    start_time).strftime('%Y-%m-%d-%H-%M-%S')

        run_mins = ceil(run_result['run_secs']/60)
        flow_name = run_result['flow_file_name']

        log_name = ('log_' + flow_name
                    + '_' + start_f
                    + '_' + str(run_mins)
                    + '.json')
        log_file = join(log_save_dir, log_name)

        # dump
        with open(log_file, "w", encoding='utf-8') as fp:
            json.dump(run_result, fp, cls=MyEncoder, sort_keys=False, indent=4)

    def _list_checkpoints(self, flow_file: str, check_file_tags: list) -> list:
        flow_info = flow_read(flow_file)
        sources = flow_info['inputs_outputs']

        check_files = []
        for source in sources:
            # get all csv outputs
            source_path = source['name']
            _, file_name = path_split(source_path)
            if source['baseType'] == 'output' and 'csv' in file_name:
                # only include file with tag
                if any(tag in file_name for tag in check_file_tags):
                    check_files.append(source_path)
        return check_files

    def list_flows(self, flow_suffix='.tfl') -> list:

        def _ignore(flow_path, ignore_tags):
            for tag in ignore_tags:
                if tag in flow_path:
                    return True
            return False

        if self._flow_pool is None:
            raise Exception("There is no flow pool setting in config")
        else:
            flow_pool_dir = self._flow_pool['pool_dir']
            ignore_tags = self._flow_pool['ignore_tags']
            _, _, files = list_files(flow_pool_dir)
            flows = []

            for file in files:
                if file.endswith(flow_suffix):
                    if ignore_tags is None:
                        flows.append(file)
                    elif _ignore(file, ignore_tags) is False:
                        flows.append(file)
            return flows

    def construct_sourcesDF(self):

        if self._flow_pool is None:
            raise Exception("There is no flow pool setting in config")
        else:
            flows = self.list_flows()
            all_sources = construct_sources(flows)

        return all_sources

    def flows_roadmap(self, target_flows, depend_flows=None) -> list:

        if depend_flows is None:
            if self._flow_pool is not None:
                depend_flows = self.list_flows()
            else:
                raise Exception("Either No depend flows OR"
                                "There is no flow pool setting in config")

        run_roadmap = flows_roadmap(target_flows=target_flows,
                                    depend_flows=depend_flows
                                    )

        return run_roadmap

    def check_checkpoints(self, flow_file) -> list:
        # set variables
        check_file_tags = self._check_file_tags

        # make list files folder path
        checkpoints = self._list_checkpoints(flow_file, check_file_tags)

        check_results = []
        if len(checkpoints) == 0:
            print(f"There is no checkpoint in the flow: {flow_file}")
        else:
            for checkpoint in checkpoints:
                checkpoint_content = read_csv(checkpoint, engine='python',
                                              encoding='utf-8-sig')
                # check
                result = checkpoint_content.shape[0] > 0
                # formated the check_result
                check_result = {'checkpoint': checkpoint,
                                'with issue': result
                                }
                check_results.append(check_result)

        return check_results

    def run_flow(self, flow_file: str) -> dict:

        max_try_times = 1 + int(self._retry_times)
        prep_script = self._assign_prep_script_byconfig(flow_file)
        credential_file = self._credential_file_byconfig(flow_file)
        check_file_tags = self._check_file_tags
        log_save_dir = self._log_save_dir

        # run
        run_results = []
        num_run = 1
        successfully_run = False
        while num_run <= max_try_times and successfully_run is False:
            run_result = flow_run(prep_script,
                                  flow_file,
                                  credential_file
                                  )
            # append info
            run_result['num_run'] = num_run
            run_result['max_try_times'] = max_try_times

            # update condition
            successfully_run = run_result['successfully_run']
            num_run += 1

            # notify run error
            if successfully_run is False:
                if self._notify_channels is not None:
                    send_msg(notify_channels=self._notify_channels,
                             msg_type='run_fail',
                             run_result=run_result
                             )
            # check checkpoints
            if successfully_run is True and check_file_tags is not None:
                check_results = self.check_checkpoints(flow_file)
                # append to run result
                run_result['check_checkpoints'] = check_results
                # get only checkpoints with issue
                checkpoints_warning = [x['checkpoint'] for x in check_results
                                       if x['with issue'] is True]

                # notify if there are some error files
                if len(checkpoints_warning) > 0:
                    send_msg(notify_channels=self._notify_channels,
                             msg_type='checkpoints_with_issue',
                             run_result=run_result,
                             checkpoints_warning=checkpoints_warning
                             )
            # save log
            if log_save_dir is not None:
                self._save_log(run_result, log_save_dir)

            # append result
            run_results.append(run_result)

        return run_results



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 24 10:34:37 2020

@author: WEIHAO
"""

from os.path import join
import subprocess
import time
from math import ceil
from prepextend.common.api import path_split
import platform

'''
examples of inputs
# prep_script_path = [Tableau Prep Builder install location]\Tableau Prep Builder <version>\scripts
# flow_file = path\\to\\[your flow file name].tfl
# credential = \\server\\path\\[your credential file name].json

flow_file_name = [your flow file name].tfl
credential_name = [your credential file name].json
flow_file_dir = path\to\
'''


def _cmd_win(prep_script: str,
             flow_file: str,
             credential_file: str = None
             ) -> str:

    # prepare the cmd which is sent to tableau prep script
    located_prep_bat = '"' + join(prep_script, 'tableau-prep-cli.bat') + '"'
    located_prep_file = ' -t ' + '"' + flow_file + '"'

    # cmd will be depend on whether there is a credential or not.
    if credential_file is not None:
        located_prep_credentials = ' -c ' + '"' + credential_file + '"'
        cmd = located_prep_bat + located_prep_credentials + located_prep_file
    else:
        cmd = located_prep_bat + located_prep_file

    # for showing
    print(cmd)

    return cmd


def _whether_run_successful(
        logs: list,
        ) -> bool:

    _success_str = 'Finished running the flow successfully'
    IsSuccess = any([_success_str.encode() in list_item for list_item in logs])

    return IsSuccess


def run(prep_script: str,
        flow_file: str,
        credential_file: str = None
        ) -> dict:

    flow_file_dir, flow_file_name = path_split(flow_file)

    os_type = platform.system()
    if os_type == 'Windows':
        cmd = _cmd_win(prep_script, flow_file, credential_file)
    else:
        raise Exception(f"Sorry, Currently We Can't Support Your OS:{os_type}")

    # run prep by prep script
    start = int(time.time())

    process = subprocess.Popen(args=cmd,
                               shell=True,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               cwd=flow_file_dir  # means cd path
                               )
    # save log
    logs = []
    for line in process.stdout:
        logs.append(line)
        print(line)

    # errcode
    errcode = process.returncode
    # show error if have
    if errcode is not None:
        raise Exception('cmd %s failed, see above for details', cmd)
    end = int(time.time())

    # produce response
    response = {'prep_script': prep_script,
                'flow_file': flow_file,
                'flow_file_dir': flow_file_dir,
                'flow_file_name': flow_file_name,
                'credential_file': credential_file,
                'os_type': os_type,
                'successfully_run': _whether_run_successful(logs),
                'start_time': start,
                'end_time': end,
                'run_secs': (end - start),
                'log': logs
                }

    return response



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 18 22:04:35 2020

@author: weiha
"""

from zipfile import ZipFile
import json
import logging as log

# flow_file = path\to\flow_file

class decode_flow:
    
    def __init__(self, flow_file = 'None'
                 ):    
        
        self.path = flow_file
        
        self.conn_class_info_dict = {'hyper':'dbname',
                                     'excel-direct':'filename'
                                    }
        self.output_types = ['csvOutputFile', 'hyperOutputFile']      
        self.input_name = 'input'
        self.output_name = 'output'
        
        if self.path != 'None':
            self.flow = self.get_flow()
        
        self.source_format = {
            'baseType':'',
            'con_type':'',
            'name':'',
            'server':'',
            'dbname':'',
            'table':'',
            'project':'',
            'datasourceName':''
            }
    
    def get_flow(self):
        # ref: https://www.coder.work/article/2033561
        log.debug(self.path)
        
        # .tlf is actually a zip file
        with ZipFile(self.path, "r") as z:
            # file 'flow' contain all detail 
            with z.open( 'flow' ) as f:  
                data = f.read()  
                data_json = json.loads(data.decode("utf-8"))
        
        return data_json
    

    def consistent_path(self, path):
        
        path = path.replace('/',"\\")
        
        return path
        
    
    def format_hyper(self, node, baseType, connection = 'None'):
        
        con_type = 'localfile'      
        if baseType == self.input_name:
            try:
                name = connection['connectionAttributes']['dbname'] # last update is in the connection
            except:
                name = node['connectionAttributes']['dbname']
         
        if baseType == self.output_name:          
            name = node['hyperOutputFile']
        
        # must with copy(), or the same items occur
        # ref: https://ithelp.ithome.com.tw/articles/10221255
        source_format = self.source_format.copy()
        source_format['baseType'] = baseType
        source_format['con_type'] = con_type
        source_format['name'] = self.consistent_path(name)
        
        return source_format 
 
    
    def format_csv(self, node, baseType, connection = 'None'):
        
        con_type = 'localfile'
        if baseType == self.input_name:
            try:
                name = connection['connectionAttributes']['filename']
            except:
                # for old version
                try:
                    name = node['connectionAttributes']['filename']
                except:
                    name = node['connectionAttributes']['directory']
         
        if baseType == self.output_name:          
            name = node['csvOutputFile']

        source_format = self.source_format.copy()
        source_format['baseType'] = baseType
        source_format['con_type'] = con_type
        source_format['name'] = self.consistent_path(name)
        
        return source_format      
    
    
    def format_excel(self, node, baseType, connection = 'None'):
        
        con_type = 'localfile'
        if baseType == self.input_name:
            name = connection['connectionAttributes']['filename']
        
        source_format = self.source_format.copy()
        source_format['baseType'] = baseType
        source_format['con_type'] = con_type
        source_format['name'] = self.consistent_path(name)
      
        return source_format


    def format_db(self, node, baseType, connection = 'None'):
        
        con_type = 'db'
        if baseType == self.input_name:
            server = connection['connectionAttributes']['server']
            # get dbname 
            dbname = connection['connectionAttributes']['dbname']
            if dbname == '':
                dbname = node['connectionAttributes']['dbname']
            # get table name
            try:
                table = node['relation']['table']
            except:
                # for custom sql query
                table = node['relation']['query']
        
        source_format = self.source_format.copy()
        source_format['baseType'] = baseType
        source_format['con_type'] = con_type
        source_format['server'] = server
        source_format['dbname'] = dbname
        source_format['table'] = table
   
        return source_format      


    def format_tableauserver(self, node, baseType, connection = 'None'):
        
        con_type =   'tableauserver'      
        if baseType == self.input_name:
            server = connection['connectionAttributes']['server'] + '/#/site/' + connection['connectionAttributes']['siteUrlName']
            project = node['connectionAttributes']['projectName']
            datasourceName = node['connectionAttributes']['datasourceName']
         
        if baseType == self.output_name:          
            server = node['serverUrl']
            project = node['projectName']
            datasourceName = node['datasourceName']

        source_format = self.source_format.copy()
        source_format['baseType'] = baseType
        source_format['con_type'] = con_type
        source_format['server'] = server
        source_format['project'] = project
        source_format['datasourceName'] = datasourceName        

        return source_format


    def list_sources(self):
                
        sources = []
        nodes = self.flow['nodes']
        connections = self.flow['connections']

        for node_id in list(nodes):
            log.debug(f"node_id: {node_id}")
            node = nodes[node_id]
            
            if node['baseType'] == 'input':
                baseType = self.input_name
                connection_id = node['connectionId']
                connection = connections[connection_id]
                log.debug(f"list_sources, input, node_id = {node}, connection_id = {connection_id}, connection = {connection}")
                # determine file_type
                try:
                    # hyper 
                    if connection['connectionAttributes']['class'] == 'hyper':
                        source_input = self.format_hyper(node, baseType, connection)
                    # csv
                    if 'LoadCsv' in node['nodeType']:
                        source_input = self.format_csv(node, baseType, connection)
                    # excel
                    if 'LoadExcel' in node['nodeType']:
                        source_input = self.format_excel(node, baseType, connection)               
                    # db
                    if (connection['connectionAttributes']['class'] == 'postgres' or 
                        connection['connectionAttributes']['class'] == 'snowflake'):
                        source_input = self.format_db(node, baseType, connection)
                    # tableauserver
                    if 'LoadSqlProxy' in node['nodeType']:
                        source_input = self.format_tableauserver(node, baseType, connection)
                        
                    sources.append(source_input)
                    
                except Exception:
                    print(f"some inputs cannot be decoded")
                    print(f"flow:{self.path}, node_id:{node_id}, baseType:{baseType}, nodetype:{node['nodeType']}")
                    raise
                    
            if node['baseType'] == 'output':
                baseType = self.output_name
                # determine file_type
                log.debug(f"list_sources, output, node_id = {node}, baseType = {baseType}")
                try:
                    # hyper 
                    if 'WriteToHyper' in node['nodeType']:
                        source_output = self.format_hyper(node, baseType)
                    # csv
                    if 'WriteToCsv' in node['nodeType']:
                        source_output = self.format_csv(node, baseType)
                    # tableauserver
                    if 'PublishExtract' in node['nodeType']:
                        source_output = self.format_tableauserver(node, baseType)  
                        
                    sources.append(source_output) 
                                              
                except Exception:
                    print(f"some outputs cannot be decoded")
                    print(f"flow:{self.path}, node_id:{node_id}, baseType:{baseType}, nodetype:{node['nodeType']}")
                    raise

        return sources


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 20 15:41:06 2020

@author: WEIHAO
"""

from strictyaml import load


class config_set:

    def __init__(self,
                 general_config,
                 version_assigned
                 ):

        self.read_config(general_config)
        self.version_assigned = self.read_version_assigned(version_assigned)

    def load_yaml(self, yaml_path):

        content = open(yaml_path, 'r', encoding="utf-8").read()
        return load(content).data

    def read_config(self, general_config):

        general_config = self.load_yaml(general_config)
        input_items = general_config.keys()

        # required config + default value
        required_items = {'default_prep_script': None,
                          'retry_times': "1",
                          'credentials_file_suffix': "_Credentials.json"
                          }

        for r_item in required_items.keys():
            key = r_item
            if r_item in input_items:
                value = general_config[key]
            else:
                value = required_items[key]
                if value is None:
                    raise Exception(f"Lack required config: {key}")
            setattr(self, key, value)
            # ref:https://stackoverflow.com/questions/52099692/how-to-assign-self-attributes-in-a-class-in-a-for-loop/52099781

        # optional config
        optional_items = ['prep_scripts_4_version_assigned',
                          'log_save_dir',
                          'check_file_tags',
                          'notify_channels',
                          'flow_pool'
                          ]

        for o_item in optional_items:
            key = o_item
            if o_item in input_items:
                value = general_config[key]
            else:
                value = None
            setattr(self, key, value)

    def read_version_assigned(self, version_assigned):

        if version_assigned is None:
            return None
        else:
            return self.load_yaml(version_assigned)



# -*- coding: utf-8 -*-
"""
Created on Wed Dec  2 13:35:57 2020

@author: WEIHAO
"""
import requests
import json


def _send_message_to_slack(webhook_url: str,
                           main_msg: str,
                           sub_msg_titles: list,
                           color_sub_msg: str,
                           ):

    # generate attachements
    attachments = []

    for sub_title in sub_msg_titles:
        attachment = {"color": color_sub_msg,
                      "title": sub_title
                      }
        attachments.append(attachment)

    # generate msg to slack api
    payload = {
        "text": main_msg,
        "attachments": attachments
        }

    # post to slack api
    response = requests.post(
        webhook_url, data=json.dumps(payload),
        headers={'Content-Type': 'application/json'}
        )


def _msg_run_fail(channel_name: str, run_result: dict) -> dict:

    msg = {}
    flow_file = run_result['flow_file']
    num_run = run_result['num_run']
    max_try_times = run_result['max_try_times']

    # generate msg for slack
    if channel_name == 'slack':
        if num_run < max_try_times:
            level = 'warning'
        else:
            level = 'danger'
        # main msg
        msg['main_msg'] = (f"[{level.upper()}] Tableau Prep Run Flow Failed. "
                           f"Run Times: {num_run}/{max_try_times}"
                           )
        # sub meg
        sub_msg = {}
        sub_msg['titles'] = [flow_file]
        sub_msg['color'] = level
        msg['sub_msg'] = sub_msg

    return msg


def _msg_checkpoints_with_issue(channel_name: str,
                                run_result: dict,
                                checkpoints_warning: list
                                ) -> dict:

    msg = {}
    flow_file = run_result['flow_file']

    # generate msg for slack
    if channel_name == 'slack':
        level = 'warning'
        # main msg
        msg['main_msg'] = (f"[{level.upper()}] Tableau Prep "
                           "Detect Some Issues in Checkpoints"
                           f"\n Flow: {flow_file}"
                           )
        # sub meg
        sub_msg = {}
        sub_msg['titles'] = checkpoints_warning
        sub_msg['color'] = level
        msg['sub_msg'] = sub_msg

    return msg


def send_msg(notify_channels: list, msg_type: str, run_result: dict,
             checkpoints_warning: list = None
             ):

    for channel in notify_channels:
        channel_name = channel['name']

        # produce msg
        if msg_type == 'run_fail':
            msg = _msg_run_fail(channel_name,
                                run_result
                                )
        if msg_type == 'checkpoints_with_issue':
            msg = _msg_checkpoints_with_issue(channel_name,
                                              run_result,
                                              checkpoints_warning
                                              )
        # send msg
        if channel_name == 'slack':
            webhook_url = channel['webhook_url']
            main_msg = msg['main_msg']
            sub_msg_titles = msg['sub_msg']['titles']
            color_sub_msg = msg['sub_msg']['color']

            _send_message_to_slack(webhook_url,
                                   main_msg,
                                   sub_msg_titles,
                                   color_sub_msg
                                   )



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 20 15:41:06 2020

@author: WEIHAO
"""

from os import walk
from os.path import join
import datetime
import json
import hashlib
from pandas import read_csv
from typing import Dict, Any
import ntpath

# In[]:


def dict_hash(dictionary: Dict[str, Any]) -> str:
    """MD5 hash of a dictionary."""
    dhash = hashlib.md5()
    # We need to sort arguments so {'a': 1, 'b': 2} is
    # the same as {'b': 2, 'a': 1}
    encoded = json.dumps(dictionary, sort_keys=True).encode()
    dhash.update(encoded)

    return dhash.hexdigest()


# In[]:

def time_utc_formated():
    time_presnet = datetime.datetime.now()
    time_now_formated = time_presnet.strftime('%Y %m %d %H %M %S')
    # format:https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior

    return time_now_formated

# In[]:


def path_split(path):
    head, tail = ntpath.split(path)
    # avoid error occur when only one layer
    return head, tail or ntpath.basename(head)


# In[]:

def list_files(folder_path):
    # include its sub-folders
    file_names = []
    file_dirs = []
    file_paths = []
    for root, dirs, files in walk(top=folder_path):
        for name in files:
            filepath = join(root, name)

            file_names.append(name)
            file_dirs.append(root)
            file_paths.append(filepath)

    return file_names, file_dirs, file_paths
    # format = series, dif vs list is serier has index
    # ref:https://discuss.analyticsvidhya.com/t/what-is-the-difference-between-pandas-series-and-python-lists/27373/2


# In[]: check_file = path\to\[your flow check name].csv

def check_fileWithRows(check_files: list):
    # make list files folder path

    error_files_name = []
    error_files_path = []
    for check_file in check_files:
        check_file_content = read_csv(check_file,
                                      engine='python',
                                      encoding='utf-8-sig')

        if check_file_content.shape[0] > 0:
            check_file_name = check_file.split('\\')[-1]

            error_files_name.append(check_file_name)
            error_files_path.append(check_file)

    return error_files_name, error_files_path



# -*- coding: utf-8 -*-
"""
Created on Wed Dec  2 13:35:57 2020

@author: WEIHAO
"""

from prepextend.run.api import (
    flow_run
    )

from prepextend.io.api import (
    flow_read
    )

from prepextend.map.api import (
    flows_roadmap,
    construct_sourcesDF
    )

from prepextend.management.api import (
    flow_manage
    )



# -*- coding: utf-8 -*-
"""
Created on Sun May  9 13:49:12 2021

@author: weiha
"""

from prepextend.io.api import flow_read
from prepextend.common.api import dict_hash
import pandas as pd
from copy import deepcopy
import logging as log
import ntpath


def _hash_sources(sources: list) -> list:

    sources_dp = deepcopy(sources)
    sources_hash = []
    for source in sources_dp:
        del source['baseType']
        source_hash = dict_hash(source)
        sources_hash.append(source_hash)

    return sources_hash


def _break_down_flows(all_sources, up_flows: list):

    # get the input_hash of up_flows
    df = all_sources.copy()
    df = df.loc[df.flow.isin(up_flows)]
    df = df.loc[df.baseType == 'input']
    input_hash = set(df.source_hash.tolist())

    # get the flows, where the output has the same hash
    df = all_sources.copy()
    df = df.loc[df.baseType == 'output']
    df = df.loc[df.source_hash.isin(input_hash)]
    down_flows = df['flow'].drop_duplicates().to_list()

    return down_flows


def _check_loop_flows(deep_nun, flow_nun, draft_flow_map):

    if deep_nun > flow_nun:  # check loop flows
        with pd.option_context('display.max_colwidth', -1):
            # for print full path
            raise Exception(f"There are some loop flows, please check"
                            f"\n"
                            f"{draft_flow_map.flow}"
                            )


def _convert(flows_roadmap) -> list:

    new_flows_roadmap = []
    for i in flows_roadmap.index.tolist():
        flow_file = flows_roadmap.loc[i].flow
        flow_dir, flow_name = ntpath.split(flow_file)

        flow_info = {'flow_file': flow_file,
                     'flow_file_dir': flow_dir,
                     'flow_file_name': flow_name
                     }

        new_flows_roadmap.append(flow_info)

    return new_flows_roadmap


def construct_sourcesDF(flows: list):

    # flows = depend_flows
    sources_df = pd.DataFrame()

    # flow = flows[0]
    for flow in flows:
        try:
            flow_info = flow_read(flow)
            sources = flow_info['inputs_outputs']

            df = pd.DataFrame(sources)
            df['flow'] = flow
            # + hash
            df.insert(0, 'source_hash',
                      _hash_sources(sources)
                      )
            # append
            sources_df = sources_df.append(df)
        except Exception:
            print(f"failed to read prep file: {flow}")

    return sources_df


def flows_roadmap(target_flows: list, depend_flows: list) -> list:

    # get all soruces
    all_sources = construct_sourcesDF(set(target_flows + depend_flows))

    # setting
    flow_nun = all_sources.flow.nunique()
    draft_flow_map = pd.DataFrame()
    up_flows = target_flows.copy()

    # make flow
    deep_nun = 0
    while len(up_flows) > 0:
        log.debug(f"up_flows: {up_flows}")
        down_flows = _break_down_flows(all_sources=all_sources,
                                       up_flows=up_flows
                                       )
        # append to whole map
        draft_flow_map = draft_flow_map.append(
            pd.DataFrame(up_flows, columns=['flow']),
            ignore_index=True
            )
        # reset the up_flows
        up_flows = down_flows.copy()
        log.debug(f"down_flows: {down_flows}")

        deep_nun += 1
        _check_loop_flows(deep_nun,
                          flow_nun,
                          draft_flow_map
                          )

    # remove duplicarte
    flows_roadmap = draft_flow_map.drop_duplicates(subset=['flow'],
                                                   keep='last',
                                                   ignore_index=True
                                                   )
    # sort
    flows_roadmap = flows_roadmap.iloc[::-1]
    flows_roadmap = flows_roadmap.reset_index(drop=True)

    # convert for better integrate with run api
    flows_roadmap = _convert(flows_roadmap)

    return flows_roadmap



# -*- coding: utf-8 -*-
"""
Created on Sun Jan 31 21:25:34 2021

@author: weiha
"""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="prepextend", # 
    version="2021.0.1",
    author="Weihao Li",
    author_email="weihao.li.tw@gmail.com",
    description="to expand the functionality of Tableau prep",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/WeihaoLiTW/prep_extension",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Environment :: Win32 (MS Windows)",
    ],
    python_requires='>=3.6',
    install_requires=["strictyaml", "requests", "pandas>=1.1.3"],
)